# line of fifths with D at origin
#
# ... Fbb Cbb Gbb Dbb Abb Ebb Bbb Fb  Cb  Gb  Db  Ab  Eb  Bb  F   C   G   D
#     -17 -16 -15 -14 -13 -12 -11 -10 -9  -8  -7  -6  -5  -4  -3  -2  -1  0
#
# D   A   E   B   F#  C#  G#  D#  A#  E#  B#  Fx  Cx  Gx  Dx  Ax  Ex  Bx  ...
# 0   +1  +2  +3  +4  +5  +6  +7  +8  +9  +10 +11 +12 +13 +14 +15 +16 +17

def natural(val):
    return abs(val) < 4

def single_sharp(val):
    return 3 < val < 11

def single_flat(val):
    return -3 > val > -11

def double_sharp(val):
    return 10 < val < 18

def double_flat(val):
    return -10 > val > -18

def modifiers(val):
    return ((val + 3) - ((val + 3) % 7)) / 7

def mod_interval(mod):
    return 7 * mod

def letter(val):
    return "DAEBFCG"[val % 7]

def name(val):
    m = modifiers(val)
    if m == 0:
        m_name = ""
    elif m == 1:
        m_name = "#"
    elif m > 1:
        m_name = "x" * (m - 1)
    else: # m < 0
        m_name = "b" * -m
    return letter(val) + m_name

def value(name):
    letter = name[0]
    base = "FCGDAEB".find(letter) - 3
    if base == -4:
        raise ValueError
    mod = name[1:]
    if mod == "":
        m = 0
    elif mod == "#":
        m = 1
    elif mod == "x" * len(mod):
        m = len(mod) + 1
    elif mod == "b" * len(mod):
        m = -len(mod)
    else:
        raise ValueError
    return base + mod_interval(m)

# tone above, new letter
def tone_above(val): 
    return val + 2

# tone below, new letter
def tone_below(val): 
    return val - 2

# semitone above, new letter
def semitone_above(val): 
    return val - 5

# semitone above, new letter
def semitone_below(val): 
    return val + 5

# semitone above, same letter
def augment(val):
    return val + 7
    
# semitone below, same latter
def diminish(val):
    return val - 7

def enharmonic(val1, val2):
    return (abs(val1 - val2) % 12) == 0

# major scale

def major_scale(tonic):
    return [tonic + i for i in [0, 2, 4, -1, 1, 3, 5]]

def minor_scale(tonic):
    return [tonic + i for i in [0, 2, -3, -1, 1, -4, -2]]


class Key:
    def __init__(self, tonic, scale):
        self.notes = scale(value(tonic))
    
    def degree_to_pitch(self, degree):
        return self.notes[degree - 1]
    
    def degree_to_pitch_and_octave(self, degree):
        o, d = divmod(degree - 1, 7)
        return self.notes[d], o


## TESTS

if __name__ == "__main__":
    assert natural(-3)
    assert natural(2)
    assert natural(0)
    assert not natural(-4)
    assert not natural(5)
    
    assert not single_sharp(0)
    assert not single_sharp(-5)
    assert single_sharp(4)
    assert not single_sharp(12)
    
    assert modifiers(0) == 0
    assert modifiers(2) == 0
    assert modifiers(-1) == 0
    assert modifiers(-5) == -1
    assert modifiers(-11) == -2
    assert modifiers(-17) == -2
    assert modifiers(4) == 1
    assert modifiers(13) == 2
    assert modifiers(17) == 2
    
    assert value("G#") == 6
    
    print [name(x) for x in major_scale(value("G#"))]



from collections import Iterable


class UnificationError(Exception):
    pass


class Point(dict):
    
    def unify(self, other):
        new = self.copy()
        for key, value in other.items():
            if key in new:
                if new[key] != value:
                    raise UnificationError(key)
            else:
                new[key] = value
        return Point(new)
    
    def tuple(self, *attributes):
        return tuple(self.get(attribute) for attribute in attributes)
    
    __mod__ = unify


class SeqBase:
    
    def __init__(self, *elements):
        if len(elements) == 1:
            if isinstance(elements[0], Point):
                elements = [elements[0]]
            elif isinstance(elements[0], Iterable):
                elements = list(elements[0])
        else:
            elements = list(elements)
        self._elements = []
        
        for point in elements:
            self.append(point)
    
    def __getitem__(self, item):
        return self._elements[item]
    
    def __len__(self):
        return len(self._elements)
    
    def __iter__(self):
        return iter(self._elements)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self._elements == other._elements
    
    def __ne__(self, other):
        return not (isinstance(other, self.__class__) and self._elements == other._elements)
    
    def map_points(self, func):
        return self.__class__([func(Point(point)) for point in self._elements])
    
    def transform(self, func):
        """
        applies function to a sequence to produce a new sequence
        """
        return func(self)
    
    __or__ = transform


def OSeq(offset_attr, duration_attr):
    
    class _OSeq(SeqBase):
        
        def last_point(self):
            if len(self._elements) == 0:
                return Point({offset_attr: 0, duration_attr: 0})
            else:
                return sorted(self._elements, key=lambda x: x[offset_attr])[-1]
        
        def next_offset(self):
            point = self.last_point()
            return point[offset_attr] + point.get(duration_attr, 0)
        
        def append(self, point):
            """
            appends a copy of the given point to this sequence, calculating
            the next offset to use for it if it doesn't have one
            """
            point = Point(point)
            if offset_attr not in point:
                point[offset_attr] = self.next_offset()
            self._elements.append(point)
        
        def concatenate(self, next_seq):
            """
            concatenates two sequences to produce a new sequence
            """
            offset = self.next_offset()
            
            new_seq = _OSeq(self._elements)
            for point in next_seq._elements:
                new_point = Point(point)
                new_point[offset_attr] = new_point[offset_attr] + offset
                new_seq._elements.append(new_point)
            return new_seq
        
        def repeat(self, count):
            """
            repeat sequence given number of times to produce a new sequence
            """
            x = _OSeq()
            for i in range(count):
                x = x.concatenate(self)
            return x
        
        def merge(self, parallel_seq):
            """
            combine the points in two sequences, putting them in offset order
            """
            return _OSeq(sorted(self._elements + parallel_seq._elements, key=lambda x: x.get(offset_attr, 0)))
        
        __add__ = concatenate
        __mul__ = repeat
        __floordiv__ = merge
    
    return _OSeq


class HSeq(SeqBase):
    """
    a horizontal sequence where each element follows the previous
    """
    
    def append(self, point):
        """
        appends a copy of the given point to this sequence
        """
        point = Point(point)
        self._elements.append(point)
    
    def concatenate(self, next_seq):
        """
        concatenates two sequences to produce a new sequence
        """
        return HSeq(self._elements + next_seq._elements)
    
    def repeat(self, count):
        """
        repeat sequence given number of times to produce a new sequence
        """
        x = HSeq()
        for i in range(count):
            x = x.concatenate(self)
        return x
    
    __add__ = concatenate
    __mul__ = repeat


class VSeq(SeqBase):
    """
    a vertical sequence where each element is coincident with the others
    """
    
    def append(self, point):
        """
        appends a copy of the given point to this sequence
        """
        point = Point(point)
        self._elements.append(point)
    
    def merge(self, parallel_seq):
        """
        combine the points in two sequences
        """
        return VSeq(self._elements + parallel_seq._elements)
    
    __floordiv__ = merge



#!/usr/bin/env python

from cStringIO import StringIO

from sebastian.core import OSequence, Point, OFFSET_64, MIDI_PITCH, DURATION_64


def write_chars(out, chars):
    out.write(chars)


def write_byte(out, b):
    out.write(chr(b))


def write_ushort(out, s):
    write_byte(out, (s >> 8) % 256)
    write_byte(out, (s >> 0) % 256)


def write_ulong(out, l):
    write_byte(out, (l >> 24) % 256)
    write_byte(out, (l >> 16) % 256)
    write_byte(out, (l >> 8) % 256)
    write_byte(out, (l >> 0) % 256)


def write_varlen(out, n):
    data = chr(n & 0x7F)
    while True:
        n = n >> 7
        if n:
            data = chr((n & 0x7F) | 0x80) + data
        else:
            break
    out.write(data)


class SMF:
    
    def __init__(self, tracks):
        self.tracks = tracks
        
    def write(self, out):
        
        num_tracks = 1 + len(self.tracks)
        Thd(format=1, num_tracks=num_tracks, division=16).write(out)
        T = 1  # how to translate events times into time_delta using the
               # division above
        
        # first track will just contain time/key/tempo info
        t = Trk()
        
        t.sequence_track_name("untitled")
        t.time_signature(4, 2, 24, 8)  # 4 4 (2nd arg is power of 2)
        t.key_signature(0, 0)  # C
        t.tempo(500000)  # in microseconds per quarter note
        t.track_end()
        t.write(out)
        
        for track in self.tracks:
            t = Trk()
            
            # we make a list of events including note off events so we can sort by
            # offset including them (to avoid negative time deltas)
            # @@@ this may eventually be a feature of sequences rather than this
            # MIDI library
            
            events_with_noteoff = []
            for point in track:
                offset, note_value, duration = point.tuple(OFFSET_64, MIDI_PITCH, DURATION_64)
                if note_value is not None:
                    events_with_noteoff.append((True, offset, note_value))
                    events_with_noteoff.append((False, offset + duration, note_value))
            
            prev_offset = None
            for on, offset, note_value in sorted(events_with_noteoff, key=lambda x: x[1]):
                if prev_offset is None:
                    time_delta = 0
                else:
                    time_delta = (offset - prev_offset) * T
                if on:
                    t.start_note(time_delta, note_value)
                else:
                    t.end_note(time_delta, note_value)
                prev_offset = offset
                
            t.track_end()
            t.write(out)


class Thd:
    
    def __init__(self, format, num_tracks, division):
        self.format = format
        self.num_tracks = num_tracks
        self.division = division
        
    def write(self, out):
        write_chars(out, "MThd")
        write_ulong(out, 6)
        write_ushort(out, self.format)
        write_ushort(out, self.num_tracks)
        write_ushort(out, self.division)


class Trk:
    
    def __init__(self):
        self.data = StringIO()
    
    def sequence_track_name(self, name):
        write_varlen(self.data, 0)  # tick
        write_byte(self.data, 0xFF)
        write_byte(self.data, 0x03)
        write_varlen(self.data, len(name))
        write_chars(self.data, name)
    
    def time_signature(self, a, b, c, d):
        write_varlen(self.data, 0)  # tick
        write_byte(self.data, 0xFF)
        write_byte(self.data, 0x58)
        write_varlen(self.data, 4)
        write_byte(self.data, a)
        write_byte(self.data, b)
        write_byte(self.data, c)
        write_byte(self.data, d)
    
    def key_signature(self, a, b):
        write_varlen(self.data, 0)  # tick
        write_byte(self.data, 0xFF)
        write_byte(self.data, 0x59)
        write_varlen(self.data, 2)
        write_byte(self.data, a)
        write_byte(self.data, b)
    
    def tempo(self, t):
        write_varlen(self.data, 0)  # tick
        write_byte(self.data, 0xFF)
        write_byte(self.data, 0x51)
        write_varlen(self.data, 3)
        write_byte(self.data, (t >> 16) % 256)
        write_byte(self.data, (t >> 8) % 256)
        write_byte(self.data, (t >> 0) % 256)
    
    def start_note(self, time_delta, note_number):
        write_varlen(self.data, time_delta)
        write_byte(self.data, 0x91)
        write_byte(self.data, note_number)
        write_byte(self.data, 64)  # velocity
    
    def end_note(self, time_delta, note_number):
        write_varlen(self.data, time_delta)
        write_byte(self.data, 0x81)
        write_byte(self.data, note_number)
        write_byte(self.data, 0)  # velocity
    
    def track_end(self):
        write_varlen(self.data, 0)  # tick
        write_byte(self.data, 0xFF)
        write_byte(self.data, 0x2F)
        write_varlen(self.data, 0)
    
    def write(self, out):
        write_chars(out, "MTrk")
        d = self.data.getvalue()
        write_ulong(out, len(d))
        out.write(d)


def write(filename, tracks):
    with open(filename, "w") as f:
        s = SMF(tracks)
        s.write(f)


if __name__ == "__main__":
    
    test = OSequence([
        Point({OFFSET_64: o, MIDI_PITCH: m, DURATION_64: d}) for (o, m, d) in [
            (0, 60, 16), (16, 72, 16), (32, 64, 16), (48, 55, 16),
            (64, 74, 16), (80, 62, 16), (96, 50, 16), (112, 48, 16),
            (128, 36, 16), (144, 24, 16), (160, 40, 16), (176, 55, 16),
            (192, 26, 16), (208, 38, 16), (224, 50, 16), (240, 48, 16)
        ]
    ])
    
    write("test.mid", [test])



from sebastian.core import MIDI_PITCH, OFFSET_64, DURATION_64
from sebastian.core import Point, OSequence

from sebastian.core.notes import modifiers


def add(properties):
    def _(point):
        point.update(properties)
        return point
    return lambda seq: seq.map_points(_)


def degree_in_key(key):
    def _(point):
        degree = point["degree"]
        pitch = key.degree_to_pitch(degree)
        point["pitch"] = pitch
        return point
    return lambda seq: seq.map_points(_)


def degree_in_key_with_octave(key, base_octave):
    def _(point):
        degree = point["degree"]
        pitch, octave = key.degree_to_pitch_and_octave(degree)
        point["pitch"] = pitch
        point["octave"] = octave + base_octave
        return point
    return lambda seq: seq.map_points(_)


def transpose(semitones):
    def _(point):
        if MIDI_PITCH in point:
            point[MIDI_PITCH] = point[MIDI_PITCH] + semitones
        return point
    return lambda seq: seq.map_points(_)


def stretch(multiplier):
    def _(point):
        point[OFFSET_64] = int(point[OFFSET_64] * multiplier)
        if DURATION_64 in point:
            point[DURATION_64] = int(point[DURATION_64] * multiplier)
        return point
    return lambda seq: seq.map_points(_)


def invert(midi_pitch_pivot):
    def _(point):
        if MIDI_PITCH in point:
            interval = point[MIDI_PITCH] - midi_pitch_pivot
            point[MIDI_PITCH] = midi_pitch_pivot - interval
        return point
    return lambda seq: seq.map_points(_)


def reverse():
    def _(sequence):
        new_elements = []
        last_offset = sequence.next_offset()
        if sequence and sequence[0][OFFSET_64] != 0:
            old_sequence = OSequence([Point({OFFSET_64: 0})]) + sequence
        else:
            old_sequence = sequence
        for point in old_sequence:
            new_point = Point(point)
            new_point[OFFSET_64] = last_offset - new_point[OFFSET_64] - new_point.get(DURATION_64, 0)
            if new_point != {OFFSET_64: 0}:
                new_elements.append(new_point)
        return OSequence(sorted(new_elements, key=lambda x: x[OFFSET_64]))
    return _


def midi_pitch():
    def _(point):
        octave = point["octave"]
        pitch = point["pitch"]
        midi_pitch = [2, 9, 4, 11, 5, 0, 7][pitch % 7]
        midi_pitch += modifiers(pitch)
        midi_pitch += 12 * octave
        point[MIDI_PITCH] = midi_pitch
        return point
    return lambda seq: seq.map_points(_)




## The goal of this module is to eventually be to MIDI players what
## 'webbrowser' is to Web browsers.
##
## Personally, I only need it to work on OS X so patches accepted for other
## operating systems.

import sys
import tempfile
import subprocess

from sebastian.midi import write_midi


def play(tracks):
    if sys.platform == "darwin":
        
        f = tempfile.NamedTemporaryFile(suffix=".mid", delete=False)
        s = write_midi.SMF(tracks)
        s.write(f)
        f.close()
        
        subprocess.call(["open", f.name])
    else:
        print "Only OS X supported at the moment. Patches accepted!"



from sebastian.core import OSequence, Point, OFFSET_64, MIDI_PITCH, DURATION_64

import re

MIDI_NOTE_VALUES = {
    "c": 0,
    "d": 2,
    "e": 4,
    "f": 5,
    "g": 7,
    "a": 9,
    "b": 11,
}


token_pattern = re.compile(r"""^\s*                 # INITIAL WHITESPACE
    (
        (                                           # NOTE
            (
                (?P<note>[abcdefg])                     # NOTE NAME
                ((?P<sharp>(is)+)|(?P<flat>(es)+)) ?    # ACCIDENTALS ?
                (?P<octave>'+|,+) ?                     # OCTAVE ?
                (=(?P<octave_check>'+|,+)) ?            # OCTAVE CHECK ?
                |                                       # or
                (?P<rest>r)                             # REST
            )
            (?P<duration>\d+\.*) ?                      # DURATION ?
            (\s*(?P<tie>~)) ?                           # TIE ?
        )
        |                                           # or
        \\(?P<command>(                             # COMMANDS
            relative | acciaccatura
        ))
        |                                           # or
        (?P<open_brace>{) | (?P<close_brace>})      # { or }
    )
    """, 
    re.VERBOSE
)


def tokenize(s):
    while True:
        if s:
            m = token_pattern.match(s)
            if m:
                yield m.groupdict()
            else:
                raise Exception("unknown token at: '%s'" % s[:20])
            s = s[m.end():]
        else:
            raise StopIteration


def note_tuple(token_dict, relative_note_tuple=None):
    note = token_dict["note"]
    octave_marker = token_dict["octave"]
    octave_check = token_dict["octave_check"]
    accidental_sharp = token_dict["sharp"]
    accidental_flat = token_dict["flat"]
    accidental_change = 0
    
    if relative_note_tuple:
        prev_note, prev_accidental_change, prev_octave = relative_note_tuple
        
        diff = MIDI_NOTE_VALUES[note] - prev_note
        if diff >= 7:
            base_octave = prev_octave - 1
        elif diff <= -7:
            base_octave = prev_octave + 1
        else:
            base_octave = prev_octave
    else:
        base_octave = 4
    
    if octave_marker is None:
        octave = base_octave
    elif "'" in octave_marker:
        octave = base_octave + len(octave_marker)
    elif "," in octave_marker:
        octave = base_octave - len(octave_marker)
    if accidental_sharp:
        accidental_change += len(accidental_sharp) / 2
    if accidental_flat:
        accidental_change -= len(accidental_flat) / 2
    
    if octave_check is None:
        pass
    elif "'" in octave_check:
        correct_octave = 4 + len(octave_check)
        if octave != correct_octave:
            print "WARNING: failed octave check" # @@@ better reporting of warning
            octave = correct_octave
    elif "," in octave_check:
        correct_octave = 4 - len(octave_check)
        if octave != correct_octave:
            print "WARNING: failed octave check" # @@@ better reporting of warning
            octave = correct_octave
    
    return MIDI_NOTE_VALUES[note], accidental_change, octave


def parse_duration(duration_marker):
    if "." in duration_marker:
        first_dot = duration_marker.find(".")
        core = int(duration_marker[:first_dot])
        # this doesn't actually check they are all dots, but regex wouldn't
        # match in the first place otherwise
        dots = len(duration_marker[first_dot:])
    else:
        core = int(duration_marker)
        dots = 0
    
    duration = int((2 - (2**-dots)) * 64 / core)
    
    return duration


def process_note(token_dict, relative_mode, prev_note_tuple):
    # @@@ there is still code duplication between here and the main parsing
    # further on
    # @@@ some of the args passed in above could be avoided if this and
    # parse_block were methods on a class
    
    duration_marker = token_dict["duration"]
    # duration must be explicit
    duration = parse_duration(duration_marker)
    
    if relative_mode:
        note_base, accidental_change, octave = note_tuple(token_dict, relative_note_tuple=prev_note_tuple)
    else:
        note_base, accidental_change, octave = note_tuple(token_dict)
    
    note_value = note_base + (12 * octave) + accidental_change
    
    return note_value, duration


def parse_block(token_generator, prev_note_tuple=None, relative_mode=False, offset=0):
    prev_duration = 16
    tie_deferred = False
    
    try:
        while True:
            token_dict = token_generator.next()
            
            command = token_dict["command"]
            open_brace = token_dict["open_brace"]
            close_brace = token_dict["close_brace"]
            
            if command:
                if command == "relative":
                    
                    token_dict = token_generator.next()
                    
                    base_note_tuple = note_tuple(token_dict)
                    
                    token_dict = token_generator.next()
                    if not token_dict["open_brace"]:
                        raise Exception("\\relative must be followed by note then {...} block")
                    
                    for obj in parse_block(token_generator, prev_note_tuple=base_note_tuple, relative_mode=True, offset=offset):
                        yield obj
                        last_offset = obj[OFFSET_64]
                    offset = last_offset
                elif command == "acciaccatura":
                    # @@@ there is much code duplication between here and the
                    # main parsing further on
                    
                    token_dict = token_generator.next()
                    note_value, duration = process_note(token_dict, relative_mode, prev_note_tuple)
                    yield Point({OFFSET_64: offset - duration / 2, MIDI_PITCH: note_value, DURATION_64: duration / 2})
                    
                    token_dict = token_generator.next()
                    note_value, duration = process_note(token_dict, relative_mode, prev_note_tuple)
                    yield Point({OFFSET_64: offset, MIDI_PITCH: note_value, DURATION_64: duration})
                    
                    offset += duration
                    prev_duration = duration
                    
                    # @@@ this should be uncommented but I'll wait until a
                    # unit test proves it should be uncommented!
                    # prev_note_tuple = note_base, accidental_change, octave
                    
            elif open_brace:
                for obj in parse_block(token_generator):
                    yield obj
            elif close_brace:
                raise StopIteration
            else:
                duration_marker = token_dict["duration"]
                rest = token_dict["rest"]
                tie = token_dict["tie"]
                
                if duration_marker is None:
                    duration = prev_duration
                else:
                    duration = parse_duration(duration_marker)
                
                if not rest:
                    if relative_mode:
                        note_base, accidental_change, octave = note_tuple(token_dict, relative_note_tuple=prev_note_tuple)
                    else:
                        note_base, accidental_change, octave = note_tuple(token_dict)
                    note_value = note_base + (12 * octave) + accidental_change
                    
                    if tie_deferred:
                        # if the previous note was deferred due to a tie
                        prev_note_value = prev_note_tuple[0] + (12 * prev_note_tuple[2]) + prev_note_tuple[1]
                        if note_value != prev_note_value:
                            raise Exception("ties are only supported between notes of same pitch")
                        duration += prev_duration
                        tie_deferred = False
                    
                    if tie:
                        # if there is a tie following this note, we defer it
                        tie_deferred = True
                    else:
                        yield Point({OFFSET_64: offset, MIDI_PITCH: note_value, DURATION_64: duration})
                    
                    prev_note_tuple = note_base, accidental_change, octave
                
                if not tie_deferred:
                    offset += duration
                
                prev_duration = duration
    except StopIteration:
        yield Point({OFFSET_64: offset})
        raise StopIteration


def parse(s, offset=0):
    return OSequence(list(parse_block(tokenize(s), offset=offset)))



# this is just an initial sketch of the data structures so don't read too 
# much into them at this stage.

# basically, a Sequence is just a collection of Points and a Point is just a
# dict giving values to certain Attributes.
#
# there are three types of Sequences: OSequences, HSequences and VSequences
# only OSequences are currently implemented
#
# OSequence assumes the Points have OFFSET_64 attribute values and
# will also make use of the DURATION_64 attribute.
#
# see datastructure_notes.txt for some of the thinking behind this whole
# approach and a bit of roadmap as to where things are headed.


OFFSET_64 = "offset_64"
MIDI_PITCH = "midi_pitch"
DURATION_64 = "duration_64"

from sebastian.core.elements import OSeq, Point, VSeq, HSeq

OSequence = OSeq(OFFSET_64, DURATION_64)

# 
# 
# def shift(offset):
#     def _(point):
#         point[OFFSET_64] = point[OFFSET_64] + offset
#         return point
#     return lambda seq: seq.map_points(_)
# 
# 



from setuptools import setup, find_packages

setup(
    name = "sebastian",
    version = "0.0.3",
    description = "symbolic music analysis and composition library in Python",
    license = "MIT",
    url = "http://github.com/jtauber/sebastian",
    author = "James Tauber",
    author_email = "jtauber@jtauber.com",
    packages = find_packages(),
)


#!/usr/bin/env python

"""
A library for parsing Standard MIDI Files (SMFs).

Currently it just outputs the data it finds.
"""


class Base:
    """
    Base is a generic base class for parsing binary files. It cannot be
    instantiated directly, you need to sub-class it and implement a parse()
    method.
    
    Your sub-class can then be instantiated with a single argument, a byte
    array.
    
    Base provides a number of basic methods for pulling data out of the byte
    array and incrementing the index appropriately.
    """
    
    def __init__(self, data):
        self.data = data
        self.index = 0
        self.parse()
    
    def peek_byte(self):
        data = self.data[self.index]
        return ord(data)
    
    def get_byte(self):
        data = self.data[self.index]
        self.index += 1
        return ord(data)
    
    def get_char(self, length=1):
        data = self.data[self.index:self.index + length]
        self.index += length
        return data
    
    def get_ushort(self):
        data = 0
        data += self.get_byte() << 8
        data += self.get_byte() << 0
        return data
    
    def get_ulong(self):
        data = 0
        data += self.get_byte() << 24
        data += self.get_byte() << 16
        data += self.get_byte() << 8
        data += self.get_byte() << 0
        return data
    
    def get_varlen(self):
        data = 0
        while True:
            next = self.get_byte()
            high_bit = next // 0x80
            data = (data << 7) + (next % 0x80)
            if not high_bit:
                return data


class SMF(Base):
    """
    A parser for Simple MIDI files.
    """
    
    def parse_chunk(self):
        chunk_id = self.get_char(4)
        length = self.get_ulong()
        data = self.get_char(length)
        return chunk_id, data
    
    def parse(self):
        while self.index < len(self.data):
            chunk_id, data = self.parse_chunk()
            if chunk_id == "MThd":
                Thd(data)
            elif chunk_id == "MTrk":
                Trk(data)
            else:
                print chunk_id
                pass  # ignore unknown chunk type


class Thd(Base):
    """
    A parser for the Thd chunk in a MIDI file.
    """
    
    def parse(self):
        format = self.get_ushort()
        num_tracks = self.get_ushort()
        division = self.get_ushort()
        print "Thd", format, num_tracks, division


class Trk(Base):
    """
    A parser for the Trk chunk in a MIDI file.
    """
    
    def process_event(self, time_delta, status):
        if status == 0xFF:
            status2 = self.get_byte()
            varlen2 = self.get_varlen()
            data = self.get_char(varlen2)
            if status2 == 0x03:
                print "sequence/track name '%s'" % data
            elif status2 == 0x04:
                print "instrument '%s'" % data
            elif status2 == 0x2F:
                assert varlen2 == 0, varlen2
                self.track_end = True
                print "track end"
            elif status2 == 0x51:
                assert varlen2 == 3, varlen2
                print "tempo %d %d %d" % (ord(data[0]), ord(data[1]), ord(data[2]))
            elif status2 == 0x54:
                assert varlen2 == 5, varlen2
                print "smpte %d %d %d %d %d" % (ord(data[0]), ord(data[1]), ord(data[2]), ord(data[3]), ord(data[4]))
            elif status2 == 0x58:
                assert varlen2 == 4, varlen2
                print "time signature %d %d %d %d" % (ord(data[0]), ord(data[1]), ord(data[2]), ord(data[3]))
            elif status2 == 0x59:
                assert varlen2 == 2, varlen2
                print "key signature %d %d" % (ord(data[0]), ord(data[1]))  # @@@ first arg signed?
            else:
                raise Exception("unknown metaevent status " + hex(status2))
        elif 0x80 <= status <= 0xEF:
            event_type, channel = divmod(status, 0x10)
            if event_type == 0x8:  # note off
                note_number = self.get_byte()
                velocity = self.get_byte()
                self.ticks += time_delta
                if self.current_note[0] == channel + 1 and self.current_note[1] == note_number:
                    pass
                else:
                    raise Exception("overlapping notes")
                start_ticks = self.current_note[2]
                beat, tick = divmod(start_ticks, 480)
                bar, beat = divmod(beat, 4)
                est_dur = divmod(time_delta + 21, 60)
                assert est_dur[1] == 0
                est_dur = est_dur[0]
                tick = divmod(tick, 60)
                assert tick[1] == 0
                tick = tick[0]
                note_name = divmod(note_number, 12)
                if self.next_note != start_ticks:
                    rest = divmod(start_ticks - self.next_note, 60)
                    assert rest[1] == 0
                    rest = rest[0]
                    print "0:0:%d" % rest,
                if bar + 1 > self.prev_bar:
                    print "\n# bar %s" % (bar + 1)
                    self.prev_bar = bar + 1
                print "%d:%d:%d" % (note_name[0], note_name[1], est_dur),
                self.next_note = start_ticks + (est_dur * 60)
            elif event_type == 0x9:  # note on
                note_number = self.get_byte()
                velocity = self.get_byte()
                self.ticks += time_delta
                self.current_note = (channel + 1, note_number, self.ticks)
                # print "event", "%s:%s:%s" % (bar + 1, beat, tick), time_delta, "note_on", channel + 1, note_number, velocity
            elif event_type == 0xB:  # controller
                controller = self.get_byte()
                value = self.get_byte()
                print "event", time_delta, "controller", channel + 1, controller, value
            elif event_type == 0xC:  # program change
                program = self.get_byte()
                print "event", time_delta, "program", channel + 1, program
            else:
                raise Exception("unknown event type " + hex(event_type))
        else:
            raise Exception("unknown status " + hex(status))
    
    def parse(self):
        self.ticks = 0
        self.next_note = 0
        self.prev_bar = 0
        global track
        track += 1
        print "\n## Track %s" % track
        self.track_end = False
        while self.index < len(self.data):
            if self.track_end:
                raise Exception("more data after track end")
            time_delta = self.get_varlen()
            next_byte = self.peek_byte()
            if next_byte >= 0x80:
                status = self.get_byte()
                self.process_event(time_delta, status)
            else:  # running status
                self.process_event(time_delta, status)  # previous status
        if not self.track_end:
            raise Exception("no track end")


if __name__ == "__main__":
    track = -1
    import sys
    filename = sys.argv[1]
    f = SMF(open(filename).read())
